#include "devices.h"

#include <stdio.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define _PATH_PROCNET_ROUTE "/proc/net/route"
#define RTF_UP          0x0001          /* route usable                 */

char *proc_gen_fmt(const char *name, int more, FILE * fh,...)
{
  char buf[512], format[512] = "";
  char *title, *head, *hdr;
  va_list ap;

  if (!fgets(buf, (sizeof buf) - 1, fh))
    return NULL;
  strcat(buf, " ");

  va_start(ap, fh);
  title = va_arg(ap, char *);
  for (hdr = buf; hdr;) {
    while (isspace(*hdr) || *hdr == '|')
      hdr++;
    head = hdr;
    hdr = strpbrk(hdr, "| \t\n");
    if (hdr)
      *hdr++ = 0;

    if (!strcmp(title, head)) {
      strcat(format, va_arg(ap, char *));
      title = va_arg(ap, char *);
      if (!title || !head)
        break;
    } else {
      strcat(format, "%*s");	/* XXX */
    }
    strcat(format, " ");
  }
  va_end(ap);

  if (!more && title) {
    fprintf(stderr, "warning: %s does not contain required field %s\n",
        name, title);
    return NULL;
  }
  return strdup(format);
}

char* rprint_fib(void)
{
  char buff[1024], iface[16];
  char gate_addr[128], net_addr[128];
  char mask_addr[128];
  int num, iflags, metric, refcnt, use, mss, window, irtt;
  FILE *fp = fopen(_PATH_PROCNET_ROUTE, "r");
  char *fmt;

  if (!fp) {
    perror(_PATH_PROCNET_ROUTE);
    printf("INET (IPv4) not configured in this system.\n");
    return 0;
  }
  irtt = 0;
  window = 0;
  mss = 0;

  fmt = proc_gen_fmt(_PATH_PROCNET_ROUTE, 0, fp,
      "Iface", "%16s",
      "Destination", "%128s",
      "Gateway", "%128s",
      "Flags", "%X",
      "RefCnt", "%d",
      "Use", "%d",
      "Metric", "%d",
      "Mask", "%128s",
      "MTU", "%d",
      "Window", "%d",
      "IRTT", "%d",
      NULL);
  /* "%16s %128s %128s %X %d %d %d %128s %d %d %d\n" */

  if (!fmt)
    return 0;

  while (fgets(buff, 1023, fp)) {
    struct sockaddr snet_target, snet_gateway, snet_mask;
    struct sockaddr_in *sin_netmask;

    num = sscanf(buff, fmt,
        iface, net_addr, gate_addr,
        &iflags, &refcnt, &use, &metric, mask_addr,
        &mss, &window, &irtt);
    if (num < 10 || !(iflags & RTF_UP))
      continue;

    if (strcmp("00000000", net_addr) == 0) {
      free(fmt);
      (void) fclose(fp);
      return strdup(iface);
    }
  }

  free(fmt);
  (void) fclose(fp);
  return 0;
}

device * determine_default_device()
{
  char *default_if;
  default_if = rprint_fib();
  if (default_if) {
    return new device(default_if);
  } else {
    return NULL;
  }
}
