/* ntm.cpp */

#include "ntm.h"

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>
#include <string>
#include <string.h>
#include <getopt.h>
#include <stdarg.h>

#include <pthread.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <pwd.h>
#include <sys/file.h>
#include <errno.h>
#include <glib/gi18n.h>
#include <locale.h>
#include <libintl.h>

#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

extern "C" {
	#include "decpcap.h"
}

#include "packet.h"
#include "connection.h"
#include "process.h"
#include "refresh.h"
#include "devices.h"

struct ntm_app {
  GtkWidget       *main_window;
  GtkWidget       *tree;
} *app;

pthread_t ntid;

extern ProcList * processes;
extern Process * unknowntcp;
extern Process * unknownudp;
extern Process * unknownip;

unsigned refreshdelay = 1;
bool tracemode = false;
bool bughuntmode = false;
bool needrefresh = true;
//packet_type packettype = packet_ethernet;
//dp_link_type linktype = dp_link_ethernet;
const char version[] = " version " VERSION "." SUBVERSION "." MINORVERSION;

const char * currentdevice = NULL;

enum
{
  COL_NAME = 0,
  COL_SENT_SEC,
  COL_RCVD_SEC,
  COL_SENT_TOTAL,
  COL_RCVD_TOTAL,
  COL_PID,
  COL_USER,
  COL_PIXBUF,
  COL_POINTER,
  NUM_COLUMNS
};

timeval curtime;
void do_refresh();
void getkbps (Process * curproc, float * recvd, float * sent);
void gettotalb(Process * curproc, float * recvd, float * sent);
char * uid2username (uid_t uid);

bool local_addr::contains (const in_addr_t & n_addr) {
	if ((sa_family == AF_INET)
	    && (n_addr == addr))
		return true;
	if (next == NULL)
		return false;
	return next->contains(n_addr);
}

bool local_addr::contains(const struct in6_addr & n_addr) {
	if (sa_family == AF_INET6)
	{
		/*
		if (DEBUG) {
			char addy [50];
			std::cerr << "Comparing: ";
			inet_ntop (AF_INET6, &n_addr, addy, 49);
			std::cerr << addy << " and ";
			inet_ntop (AF_INET6, &addr6, addy, 49);
			std::cerr << addy << std::endl;
		}
		*/
		//if (addr6.s6_addr == n_addr.s6_addr)
		if (memcmp (&addr6, &n_addr, sizeof(struct in6_addr)) == 0)
		{
			if (DEBUG)
				std::cerr << "Match!" << std::endl;
			return true;
		}
	}
	if (next == NULL)
		return false;
	return next->contains(n_addr);
}

struct dpargs {
	int sa_family;
	in_addr ip_src;
	in_addr ip_dst;
	in6_addr ip6_src;
	in6_addr ip6_dst;
};

int process_tcp (u_char * userdata, const dp_header * header, const u_char * m_packet) {
	struct dpargs * args = (struct dpargs *) userdata;
	struct tcphdr * tcp = (struct tcphdr *) m_packet;

	curtime = header->ts;

	/* get info from userdata, then call getPacket */
	Packet * packet;
	switch (args->sa_family)
	{
		case (AF_INET):
			packet = new Packet (args->ip_src, ntohs(tcp->source), args->ip_dst, ntohs(tcp->dest), header->len, header->ts);
			break;
		case (AF_INET6):
			packet = new Packet (args->ip6_src, ntohs(tcp->source), args->ip6_dst, ntohs(tcp->dest), header->len, header->ts);
			break;
	}

	Connection * connection = findConnection(packet);

	if (connection != NULL)
	{
		/* add packet to the connection */
		connection->add(packet);
	} else {
		/* else: unknown connection, create new */
		connection = new Connection (packet);
		getProcess(connection, currentdevice);
	}
	delete packet;

	if (needrefresh)
	{
		do_refresh();
		needrefresh = false;
	}

	/* we're done now. */
	return true;
}

int process_udp (u_char * userdata, const dp_header * header, const u_char * m_packet) {
	struct dpargs * args = (struct dpargs *) userdata;
	//struct tcphdr * tcp = (struct tcphdr *) m_packet;
	struct udphdr * udp = (struct udphdr *) m_packet;

	curtime = header->ts;

	/* TODO get info from userdata, then call getPacket */
	Packet * packet;
	switch (args->sa_family)
	{
		case (AF_INET):
			packet = new Packet (args->ip_src, ntohs(udp->source), args->ip_dst, ntohs(udp->dest), header->len, header->ts);
			break;
		case (AF_INET6):
			packet = new Packet (args->ip6_src, ntohs(udp->source), args->ip6_dst, ntohs(udp->dest), header->len, header->ts);
			break;
	}

	//if (DEBUG)
	//	std::cout << "Got packet from " << packet->gethashstring() << std::endl;

	Connection * connection = findConnection(packet);

	if (connection != NULL)
	{
		/* add packet to the connection */
		connection->add(packet);
	} else {
		/* else: unknown connection, create new */
		connection = new Connection (packet);
		getProcess(connection, currentdevice);
	}
	delete packet;

	if (needrefresh)
	{
		do_refresh();
		needrefresh = false;
	}

	/* we're done now. */
	return true;
}

int process_ip (u_char * userdata, const dp_header * /* header */, const u_char * m_packet) {
	struct dpargs * args = (struct dpargs *) userdata;
	struct ip * ip = (struct ip *) m_packet;
	args->sa_family = AF_INET;
	args->ip_src = ip->ip_src;
	args->ip_dst = ip->ip_dst;

	/* we're not done yet - also parse tcp :) */
	return false;
}

int process_ip6 (u_char * userdata, const dp_header * /* header */, const u_char * m_packet) {
	struct dpargs * args = (struct dpargs *) userdata;
	const struct ip6_hdr * ip6 = (struct ip6_hdr *) m_packet;
	args->sa_family = AF_INET6;
	args->ip6_src = ip6->ip6_src;
	args->ip6_dst = ip6->ip6_dst;

	/* we're not done yet - also parse tcp :) */
	return false;
}

void quit_cb (int /* i */)
{
	procclean();
	exit(0);
}

void forceExit(const char *msg, ...)
{
	va_list argp;
	va_start(argp, msg);
	vfprintf(stderr, msg, argp);
	va_end(argp);
	std::cerr << std::endl;

    exit(0);
}

class handle {
public:
	handle (dp_handle * m_handle, const char * m_devicename = NULL,
			handle * m_next = NULL) {
		content = m_handle; next = m_next; devicename = m_devicename;
	}
	dp_handle * content;
	const char * devicename;
	handle * next;
};

void
list_store_update(GtkTreeModel *model, GtkTreeIter *iter, int column, const gulong new_value)
{
  gulong current_value;
  gtk_tree_model_get(model, iter, column, &current_value, -1);
  if (current_value != new_value) {
    gtk_list_store_set(GTK_LIST_STORE(model), iter, column, new_value, -1);
  }
}

static void
update_info_mutable_cols(Process *info)
{
  float value_sent = 0;
  float value_recv = 0;
  float total_sent = 0;
  float total_recv = 0;
  GtkTreeModel *model;

  getkbps (info, &value_recv, &value_sent);
  gettotalb(info, &total_recv, &total_sent);

  model = gtk_tree_view_get_model(GTK_TREE_VIEW(app->tree));

  list_store_update(model, &info->iter, COL_SENT_SEC, value_sent);
  list_store_update(model, &info->iter, COL_RCVD_SEC, value_recv);
  list_store_update(model, &info->iter, COL_SENT_TOTAL, total_sent);
  list_store_update(model, &info->iter, COL_RCVD_TOTAL, total_recv);
}

static void
insert_info_to_tree(Process *info)
{
  GtkTreeModel *model;
  uid_t uid = info->getUid();

  model = gtk_tree_view_get_model(GTK_TREE_VIEW(app->tree));

  gtk_list_store_insert(GTK_LIST_STORE(model), &info->iter, 0);
  info->in_list = true;

  gtk_list_store_set(GTK_LIST_STORE(model), &info->iter,
      COL_NAME, strdup(info->name),
      COL_PID, info->pid,
      COL_USER, uid2username(uid),
      -1);
}

static void
remove_info_from_tree(GtkTreeModel *model, Process *info)
{
  gtk_list_store_remove(GTK_LIST_STORE(model), &info->iter);
  info->in_list = false;
}

static guint64
get_size_from_column(GtkTreeModel *model, GtkTreeIter *first,
    const guint index)
{
  GValue value = G_VALUE_INIT;
  gtk_tree_model_get_value(model, first, index, &value);

  guint64 size;
  if (G_VALUE_TYPE(&value) == G_TYPE_ULONG) {
    size = g_value_get_ulong(&value);
  } else {
    g_assert_not_reached();
  }
  g_value_unset(&value);
  return size;
}

extern "C" {
  G_MODULE_EXPORT gboolean
    cb_main_window_delete(GtkWidget *, GdkEvent *, gpointer)
    {
      gtk_main_quit();
      return true;
    }

  G_MODULE_EXPORT void
    bytes_cell_data_func(GtkTreeViewColumn *, GtkCellRenderer *renderer,
        GtkTreeModel *model, GtkTreeIter *iter,
        gpointer user_data)
    {
      const guint index = GPOINTER_TO_UINT(user_data);
      guint64 size;
      GValue value = G_VALUE_INIT;
      gtk_tree_model_get_value(model, iter, index, &value);
      
      if (G_VALUE_TYPE(&value) == G_TYPE_ULONG) {
        size = g_value_get_ulong(&value);
      } else {
        g_assert_not_reached();
      }
      g_value_unset(&value);

      if (size == 0) {
        char *str = g_strdup_printf("<i>%s</i>", _("N/A"));
        g_object_set(renderer, "markup", str, NULL);
        g_free(str);
      } else {
        char *str = g_format_size_full(size, G_FORMAT_SIZE_IEC_UNITS);
        g_object_set(renderer, "text", str, NULL);
        g_free(str);
      }
    }
  G_MODULE_EXPORT void
    cb_row_selected(GtkTreeSelection *, gpointer)
    {
    }
  G_MODULE_EXPORT void
    cb_refresh_icons(GtkIconTheme *, gpointer)
    {
    }
  G_MODULE_EXPORT gint
    number_compare_func(GtkTreeModel *model, GtkTreeIter *first,
        GtkTreeIter *second, gpointer user_data)
    {
      const guint index = GPOINTER_TO_UINT(user_data);

      guint64 size1, size2;
      size1 = get_size_from_column(model, first, index);
      size2 = get_size_from_column(model, second, index);
      if (size2 > size1) {
        return 1;
      } else if (size2 < size1) {
        return -1;
      }
      return 0;
    }
}

void *
thr_fn(void *)
{
	process_init();

	device * devices = NULL;
	//dp_link_type linktype = dp_link_ethernet;
	int promisc = 0;

  devices = determine_default_device();

	if (NEEDROOT && (geteuid() != 0))
		forceExit("You need to be root to run ntm!");

	char errbuf[PCAP_ERRBUF_SIZE];

	handle * handles = NULL;
	device * current_dev = devices;
	while (current_dev != NULL) {
		getLocal(current_dev->name, tracemode);
		if ((!tracemode) && (!DEBUG)){
			//caption->append(current_dev->name);
			//caption->append(" ");
		}

		dp_handle * newhandle = dp_open_live(current_dev->name, BUFSIZ, promisc, 100, errbuf);
		if (newhandle != NULL)
		{
			dp_addcb (newhandle, dp_packet_ip, process_ip);
			dp_addcb (newhandle, dp_packet_ip6, process_ip6);
			dp_addcb (newhandle, dp_packet_tcp, process_tcp);
			dp_addcb (newhandle, dp_packet_udp, process_udp);

			/* The following code solves sf.net bug 1019381, but is only available
			 * in newer versions (from 0.8 it seems) of libpcap
			 *
			 * update: version 0.7.2, which is in debian stable now, should be ok
			 * also.
			 */
			if (dp_setnonblock (newhandle, 1, errbuf) == -1)
			{
				fprintf(stderr, "Error putting libpcap in nonblocking mode\n");
			}
			handles = new handle (newhandle, current_dev->name, handles);
		}
		else
		{
			fprintf(stderr, "Error opening handler for device %s\n", current_dev->name);
		}

		current_dev = current_dev->next;
	}

	signal (SIGALRM, &alarm_cb);
	signal (SIGINT, &quit_cb);
	alarm (refreshdelay);

	// Main loop:
	//
	//  Walks though the 'handles' list, which contains handles opened in non-blocking mode.
	//  This causes the CPU utilisation to go up to 100%. This is tricky:
	while (1)
	{
		bool packets_read = false;

		handle * current_handle = handles;
		while (current_handle != NULL)
		{
			struct dpargs * userdata = (dpargs *) malloc (sizeof (struct dpargs));
			userdata->sa_family = AF_UNSPEC;
			currentdevice = current_handle->devicename;
			int retval = dp_dispatch (current_handle->content, -1, (u_char *)userdata, sizeof (struct dpargs));
			if (retval == -1 || retval == -2)
			{
				std::cerr << "Error dispatching" << std::endl;
			}
			else if (retval != 0)
			{
				packets_read = true;
			}
			free (userdata);
			current_handle = current_handle->next;
		}

		if (needrefresh)
		{
			do_refresh();
			needrefresh = false;
		}

		// If no packets were read at all this iteration, pause to prevent 100%
		// CPU utilisation;
		if (!packets_read)
		{
			usleep(100);
		}
	}
  return ((void *)0);
}

char * uid2username (uid_t uid)
{
	struct passwd * pwd = NULL;
	/* getpwuid() allocates space for this itself,
	 * which we shouldn't free */
	pwd = getpwuid(uid);

	if (pwd == NULL)
	{
		assert(false);
		return strdup ("unlisted");
	} else {
		return strdup(pwd->pw_name);
	}
}

float tobps (u_int32_t bytes)
{
	return ((double)bytes) / PERIOD;
}

/** Get the kb/s values for this process */
void getkbps (Process * curproc, float * recvd, float * sent)
{
	u_int32_t sum_sent = 0,
	  	sum_recv = 0;

	/* walk though all this process's connections, and sum
	 * them up */
	ConnList * curconn = curproc->connections;
	ConnList * previous = NULL;
	while (curconn != NULL)
	{
		if (curconn->getVal()->getLastPacket() <= curtime.tv_sec - CONNTIMEOUT)
		{
			/* stalled connection, remove. */
			ConnList * todelete = curconn;
			Connection * conn_todelete = curconn->getVal();
			curconn = curconn->getNext();
			if (todelete == curproc->connections)
				curproc->connections = curconn;
			if (previous != NULL)
				previous->setNext(curconn);
			delete (todelete);
			delete (conn_todelete);
		}
		else
		{
			u_int32_t sent = 0, recv = 0;
			curconn->getVal()->sumanddel(curtime, &recv, &sent);
			sum_sent += sent;
			sum_recv += recv;
			previous = curconn;
			curconn = curconn->getNext();
		}
	}
	*recvd = tobps(sum_recv);
	*sent = tobps(sum_sent);
}

/** get total values for this process */
void gettotal(Process * curproc, u_int32_t * recvd, u_int32_t * sent)
{
	u_int32_t sum_sent = 0,
	  	sum_recv = 0;
	ConnList * curconn = curproc->connections;
	while (curconn != NULL)
	{
		Connection * conn = curconn->getVal();
		sum_sent += conn->sumSent;
		sum_recv += conn->sumRecv;
		curconn = curconn->getNext();
	}
	//std::cout << "Sum sent: " << sum_sent << std::endl;
	//std::cout << "Sum recv: " << sum_recv << std::endl;
	*recvd = sum_recv;
	*sent = sum_sent;
}

void gettotalb(Process * curproc, float * recvd, float * sent)
{
	u_int32_t sum_sent = 0,
	  	sum_recv = 0;
	gettotal(curproc, &sum_recv, &sum_sent);
	//std::cout << "Total sent: " << sum_sent << std::endl;
	*sent = sum_sent;
	*recvd = sum_recv;
}

gboolean
update_liststore(gpointer user_data)
{
	ProcList * curproc = processes;
	ProcList * previousproc = NULL;
	int nproc = processes->size();
	int n = 0;
  GtkTreeModel *model;

  model = gtk_tree_view_get_model(GTK_TREE_VIEW(app->tree));

	while (curproc != NULL)
	{
		// walk though its connections, summing up their data, and
		// throwing away connections that haven't received a package
		// in the last PROCESSTIMEOUT seconds.
		assert (curproc != NULL);
		assert (curproc->getVal() != NULL);
		assert (nproc == processes->size());

		/* remove timed-out processes (unless it's one of the the unknown process) */
		if ((curproc->getVal()->getLastPacket() + PROCESSTIMEOUT <= curtime.tv_sec)
				&& (curproc->getVal() != unknowntcp)
				&& (curproc->getVal() != unknownudp)
				&& (curproc->getVal() != unknownip))
		{
			if (DEBUG)
				std::cout << "PROC: Deleting process\n";
			ProcList * todelete = curproc;
			Process * p_todelete = curproc->getVal();
			if (previousproc)
			{
				previousproc->next = curproc->next;
				curproc = curproc->next;
			} else {
				processes = curproc->getNext();
				curproc = processes;
			}
      remove_info_from_tree(model, p_todelete);
			delete todelete;
			delete p_todelete;
			nproc--;
			//continue;
		}
		else
		{
			// add a non-timed-out process to the list of stuff to show
      if (!curproc->getVal()->in_list) {
        insert_info_to_tree(curproc->getVal());
      }
      update_info_mutable_cols(curproc->getVal());

			previousproc = curproc;
			curproc = curproc->next;
			n++;
		}
	}
  return false;
}

// Display all processes and relevant network traffic using show function
void do_refresh()
{
  gdk_threads_add_idle((GSourceFunc)update_liststore, (gpointer)processes);

	if (tracemode || DEBUG) {
		/* print the 'unknown' connections, for debugging */
		ConnList * curr_unknownconn = unknowntcp->connections;
		while (curr_unknownconn != NULL) {
			std::cout << "Unknown connection: " <<
				curr_unknownconn->getVal()->refpacket->gethashstring() << std::endl;

			curr_unknownconn = curr_unknownconn->getNext();
		}
	}
}

GtkWidget *
proctable_new()
{
  GtkWidget *proctree;
  GtkListStore *model;
  GtkTreeSelection *selection;
  GtkTreeViewColumn *column;
  GtkCellRenderer *cell_renderer;
  gint i;

  const gchar *titles[] = {
    N_("Process Name"),
    N_("Sent/sec"),
    N_("Received/sec"),
    N_("Sent"),
    N_("Received"),
    N_("PID"),
    N_("User"),
    NULL,
    "POINTER"
  };

  model = gtk_list_store_new(NUM_COLUMNS,
      G_TYPE_STRING,  /* Process Name */
      G_TYPE_ULONG,   /* Sent/sec */
      G_TYPE_ULONG,   /* Received/sec */
      G_TYPE_ULONG,   /* Sent */
      G_TYPE_ULONG,   /* Received */
      G_TYPE_UINT,    /* PID */
      G_TYPE_STRING,  /* User */
      GDK_TYPE_PIXBUF,/* Icon */
      G_TYPE_POINTER  /* ProcInfo */
      );

  proctree = gtk_tree_view_new_with_model((GTK_TREE_MODEL(model)));
  gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(proctree), TRUE);
  g_object_unref(G_OBJECT(model));

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(proctree));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

  column = gtk_tree_view_column_new();

  cell_renderer = gtk_cell_renderer_pixbuf_new();
  gtk_tree_view_column_pack_start(column, cell_renderer, FALSE);
  gtk_tree_view_column_set_attributes(column, cell_renderer,
      "pixbuf", COL_PIXBUF,
      NULL);

  cell_renderer = gtk_cell_renderer_text_new();
  gtk_tree_view_column_pack_start(column, cell_renderer, FALSE);
  gtk_tree_view_column_set_attributes(column, cell_renderer,
      "text", COL_NAME,
      NULL);
  gtk_tree_view_column_set_title(column, _(titles[0]));

  gtk_tree_view_column_set_sort_column_id(column, COL_NAME);
  gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
  gtk_tree_view_column_set_reorderable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(proctree), column);
  gtk_tree_view_set_expander_column(GTK_TREE_VIEW(proctree), column);

  for (i = COL_SENT_SEC; i <= COL_USER; i++) {
    GtkCellRenderer *cell;
    GtkTreeViewColumn *col;

    cell = gtk_cell_renderer_text_new();
    col = gtk_tree_view_column_new();
    gtk_tree_view_column_pack_start(col, cell, TRUE);
    gtk_tree_view_column_set_title(col, _(titles[i]));
    gtk_tree_view_column_set_sort_column_id(col, i);
    gtk_tree_view_column_set_sizing(col, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
    gtk_tree_view_column_set_reorderable(col, TRUE);
    gtk_tree_view_append_column(GTK_TREE_VIEW(proctree), col);

    switch (i) {
      case COL_SENT_SEC:
      case COL_RCVD_SEC:
      case COL_SENT_TOTAL:
      case COL_RCVD_TOTAL:
        gtk_tree_view_column_set_cell_data_func(col, cell,
            bytes_cell_data_func,
            GUINT_TO_POINTER(i),
            NULL);
        break;
      default:
        gtk_tree_view_column_set_attributes(col, cell, "text", i, NULL);
        break;
    }

    switch (i) {
      case COL_SENT_SEC:
      case COL_RCVD_SEC:
      case COL_SENT_TOTAL:
      case COL_RCVD_TOTAL:
        gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(model), i,
            number_compare_func, GUINT_TO_POINTER(i),
            NULL);
        break;
      default:
        break;
    }
    
    switch (i) {
      case COL_SENT_SEC:
      case COL_RCVD_SEC:
      case COL_SENT_TOTAL:
      case COL_RCVD_TOTAL:
      case COL_PID:
        g_object_set(G_OBJECT(cell), "xalign", 1.0f, NULL);
        break;
      default:
        break;
    }
  }
  GtkIconTheme *theme = gtk_icon_theme_get_default();
  g_signal_connect(G_OBJECT(theme), "changed", G_CALLBACK(cb_refresh_icons), NULL);
  g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(proctree))),
      "changed",
      G_CALLBACK(cb_row_selected), NULL);

  return proctree;
}

int
main (int argc, char** argv)
{
  int err;
  int pid_file;
  int rc;
  GtkBuilder *builder;
  GtkWidget *dialog, *scrolled;

  // Set the current local to default
  setlocale(LC_ALL, "");

  // bindtextdomain(DOMAINNAME, DIRNAME)
  //
  // Specify that the DOMAINNAME message catalog
  // will be found in DIRNAME rather than in 
  // the system locale data base.
  bindtextdomain("ntm", "/usr/share/locale");

  // testdomain(DOMAINNAME)
  //
  // Set the current default message catalog to DOMAINNAME.
  textdomain("ntm");

  gtk_init(&argc, &argv);
  builder = gtk_builder_new();

  pid_file = open("/var/run/ntm.pid", O_CREAT | O_RDWR, 0666);
  rc = flock(pid_file, LOCK_EX | LOCK_NB);
  if (rc) {
    if (EWOULDBLOCK == errno) {
      gtk_builder_add_from_file(builder, "/usr/lib/ntm/data/ui/alart.glade", NULL);
      dialog = GTK_WIDGET(gtk_builder_get_object(builder, "dialog"));
      gtk_builder_connect_signals(builder, NULL);
      gtk_widget_show(dialog);
    }
  } else {
    app = (struct ntm_app *) malloc(sizeof(struct ntm_app));
    err = pthread_create(&ntid, NULL, thr_fn, NULL);
    if (err != 0) {
      printf("can't create thread: %s\n", strerror(err));
    }
    gtk_builder_add_from_file(builder, "/usr/lib/ntm/data/ui/ntm.glade", NULL);
    app->main_window = GTK_WIDGET(gtk_builder_get_object(builder, "main_window"));
    app->tree = proctable_new();
    scrolled = GTK_WIDGET(gtk_builder_get_object(builder, "processes_scrolled"));
    gtk_container_add(GTK_CONTAINER(scrolled), app->tree);
    g_signal_connect(G_OBJECT(app->main_window), "delete_event",
        G_CALLBACK(cb_main_window_delete), NULL);
    gtk_widget_show_all(app->main_window);
  }

  gtk_main();

  if (app) {
    free(app);
  }

  exit(0);
}
